/*
 * Copyright (c) 2006
 *
 * written by Simon Barrick
 * simon@fourthwall.ndo.co.uk
 *
 * PolyCrease SOP
 */


#ifndef __SOP_PolyCrease_h__
#define __SOP_PolyCrease_h__

#include <SOP/SOP_Node.h>
#include <GEO/GEO_PointTree.h>
#include <GB/GB_EdgeGroup.h>
#include <GEO/GEO_Closure.h>
#include <GU/GU_EdgeSplit.h>


class SOP_PolyCrease : public SOP_Node
{
public:
       SOP_PolyCrease(OP_Network *net, const char *name, OP_Operator *op);
    virtual ~SOP_PolyCrease();

    static PRM_Template    myTemplateList[];
    static OP_Node    *myConstructor(OP_Network*, const char *,
                  OP_Operator *);

protected:

    // Method to cook geometry for the SOP
    virtual unsigned      disableParms();
    virtual OP_ERROR      cookMySop(OP_Context &context);
    virtual const char    *inputLabel(unsigned int) const;

private:
    bool checkFlipEdge(GEO_Point *pta0, GEO_Point *ptb0, float t, GEO_Point *&pta, GEO_Point *&ptb, float &u);
    void  getGroups(UT_String &str) { evalString(str, 0, 0, 0); }

    //parameters
    void  outGroup(UT_String &str, UT_String chan)  { evalString(str, chan, 0, 0); }
    bool  CLOSEPATHS()          { return evalInt("closepaths", 0, 0); }
    bool  NEWT()                { return evalInt("newt", 0, 0); }
    bool  SIDERAILS()           { return evalInt("siderails", 0, 0); }
    bool  USEBIAS()             { return evalInt("newt", 0, 0); }
    float BIAS(float t)         { return evalFloat("t", 0, t); }
    bool  PTYPE()               { return evalInt("pathtype", 0, 0); }
    int   EMBED()               { return evalInt("embed", 0, 0); }
    bool  UPDATE()              { return evalInt("norms", 0, 0); }
    float TOL(float t)          { return evalFloat("tol", 0, t); }
    bool  OUTGRP()              { return evalInt("outgrps", 0, 0); }
    bool  ENDPOINTS()           { return evalInt("snapendrails", 0, 0); }
    bool  NEWANG()              { return evalInt("newang", 0, 0); }
    bool  PARA()                { return evalInt("disttype", 0, 0); }
    bool  PARTITION()           { return evalInt("partition", 0, 0); }
    int   SEED(float t)         { return evalInt("seed", 0, t); }
    float OFFSET(UT_String chan,float t)       { return evalFloat(chan, 0, t); }
    void  PREFIX(UT_String &str)  { evalString(str, "prefix", 0, 0); }


    bool findEdgeUsingPoint(float &u, GEO_Point *&pta, GEO_Point *&ptb, GEO_Primitive *&prim,
                            UT_Vector3 ppt, const GEO_Point *nearppt,
                            const GB_PointGroup *ptgroup, GB_PointRefArray *prefarr, GB_PrimitiveGroup *primlist, float thres);

    bool findEdgeUsingPrim(float &u, GEO_Point *&pta, GEO_Point *&ptb, GEO_Primitive *&prim,
                           UT_Vector3 ppt,  GB_PrimitiveGroup *primlist, float thres);

    void appendSplitLocs(float now, bool found, float u, GEO_Point *pta, GEO_Point *ptb, GEO_Primitive *prim);

    void createExtraSplitLocs(float now);

    void makeGroups(GU_PointGroup &newGroup, float tol);

    void huntGdp(float &u, GEO_Point *&pta, GEO_Point *&ptb, GEO_Primitive *&prim,
                 UT_Vector3 ppt, GB_PrimitiveGroup *primlist, GB_PointRefArray *prefarr, GB_PointGroup *ptgroup,
                 float tol, bool &found);

    // This variable is used together with the call to the "checkInputChanged"
    // routine to notify the handles (if any) if the input has changed.
    GU_DetailGroupPair   myDetailGroupPair;

    // This is the group of geometry to be manipulated by this SOP and cooked
    // by the method "cookInputGroups".
    GB_PointGroup *myGroup;
    UT_String  grp_name;

    //data
    GU_SplitLocPtrArray splitLocs;
    GU_SplitLocPtrArray allSplitLocs;
    GU_SplitLocPtrArray scratchSplitLocs;
    int currentSplitCount;
    GU_Detail *scratchgdp;
    GU_Detail *refgdp;
    GEO_PointTree scratchpointtree;
    GEO_PointTree pointtree;
    GB_EdgeGroup *scratchoutedges;
    GB_EdgeGroup *outedges;
    GEO_Closure *scratchConnectedPrims;
    GEO_Closure *connectedPrims;

};

#endif
