/*
 * Copyright (c) 2006
 *
 * written by Simon Barrick
 * simon@fourthwall.ndo.co.uk
 *
 * PolyCrease SOP
 */

#define MAXFLT FLT_MAX
#define USEPOLYS 1

#include <UT/UT_DSOVersion.h>
#include <UT/UT_Math.h>
#include <PRM/PRM_Include.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include <UT/UT_Interrupt.h>

#include <GU/GU_PointGroup.h>
#include <GU/GU_Group.h>
#include <SOP/SOP_Node.h>
#include <GB/GB_Group.h>
#include <GEO/GEO_Closure.h>
#include <GEO/GEO_TPSurf.h>
#include <GB/GB_PointRef.h>
#include <GEO/GEO_PointTree.h>
#include <GU/GU_EdgeSplit.h>
#include <UT/UT_BoundingBox.h>
#include <UT/UT_IntArray.h>
#include <GB/GB_EdgeGroup.h>

#include "SOP_PolyCrease.h"

void
newSopOperator(OP_OperatorTable *table)
{
     table->addOperator(new OP_Operator("PolyCrease",
          "PolyCrease",
           SOP_PolyCrease::myConstructor,
           SOP_PolyCrease::myTemplateList,
           1,
           2,
           0));
}

static PRM_Name embedMenuNames[] =
{
  PRM_Name( "none", "No embedding" ),

  PRM_Name( "extract", "Extract and destroy" ),

  PRM_Name( "overwrite", "Create with overwrite" ),

  //PRM_Name( "append", "Append" ),

  //PRM_Name( "merge", "Merge" ),

  PRM_Name(0)

};

static PRM_ChoiceList   embedMenu((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE
                                                    | PRM_CHOICELIST_REPLACE),
                                                      embedMenuNames);

static PRM_Name typeMenuNames[] =
{
  PRM_Name( "short", "Shortest path" ),

  PRM_Name( "quad", "Quad strips" ),

  PRM_Name(0)

};

static PRM_ChoiceList   typeMenu((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE
                                                    | PRM_CHOICELIST_REPLACE),
                                                      typeMenuNames);
//
static PRM_Name distMenuNames[] =
{
  PRM_Name( "actual", "Actual" ),

  PRM_Name( "para", "Parametric" ),

  PRM_Name(0)

};

static PRM_ChoiceList   distMenu((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE
                                                    | PRM_CHOICELIST_REPLACE),
                                                      distMenuNames);

static PRM_Name        names[] = {
  //0
    PRM_Name("pathtype",  "Path type"),
  //1
    PRM_Name("newt",  "Override bias"),
  //2
    PRM_Name("t",  "Bias"),
  //3
    PRM_Name("norms",  "Update point normals"),
  //4
    PRM_Name("closepaths",  "Close paths"),
  //5
    PRM_Name("siderails",  "Side rails"),
  //6
    PRM_Name("topoffset",  "Top rail offset"),
  //7
    PRM_Name("botoffset",  "Bot rail offset"),
  //8
    PRM_Name("embed",  "Embed creases"),
  //9
    PRM_Name("tol",  "Tolerance"),
  //10
    PRM_Name("outgrps",  "Output groups"),
  //11
    PRM_Name("agrp",  "A group"),
  //12
    PRM_Name("ongrp",  "Group name"),
  //13
    PRM_Name("bgrp",  "B group"),
  //14
    PRM_Name("snapendrails",  "Snap siderails to end points"),
  //15
   PRM_Name("newang",  "Override angle"),
  //16
   PRM_Name("ang",  "Angle"),
   //17
   PRM_Name("disttype",  "Dist type"),
   //18
   PRM_Name("partition",  "Partition groups"),
   //19
   PRM_Name("seed",  "Hint point"),
   //20
   PRM_Name("prefix",  "Attribute Prefix"),


};

static PRM_Default PRMtolDefault(0.001f);
static PRM_Default PRMinGrpDefault(0.0f, "Agroup");
static PRM_Default PRMonGrpDefault(0.0f, "onCrease");
static PRM_Default PRMoutGrpDefault(0.0f, "Bgroup");
static PRM_Default PRMprefixDefault(0.0f, "");

PRM_Template
SOP_PolyCrease::myTemplateList[] = {
    //PRM_Template(PRM_STRING,    1, &PRMgroupName, 0, &SOP_Node::pointGroupMenu),

    PRM_Template(PRM_ORD, 1, &names[8], PRMzeroDefaults, &embedMenu),

    PRM_Template(PRM_ORD, 1, &names[0], PRMzeroDefaults, &typeMenu),

    PRM_Template(PRM_TOGGLE, 1, &names[1], PRMzeroDefaults),

    PRM_Template(PRM_FLT_J, 1, &names[2], PRMpointFiveDefaults),

    PRM_Template(PRM_TOGGLE, 1, &names[15], PRMzeroDefaults),

    PRM_Template(PRM_FLT_J, 1, &names[16], PRM90Defaults),

    PRM_Template(PRM_TOGGLE, 1, &names[3], PRMoneDefaults),

    PRM_Template(PRM_TOGGLE, 1, &names[4], PRMzeroDefaults),

    PRM_Template(PRM_TOGGLE, 1, &names[5], PRMzeroDefaults),

    PRM_Template(PRM_ORD, 1, &names[17], PRMoneDefaults, &distMenu),

    PRM_Template(PRM_TOGGLE, 1, &names[14], PRMzeroDefaults),

    PRM_Template(PRM_FLT_J, 1, &names[6], PRMpointOneDefaults),

    PRM_Template(PRM_FLT_J, 1, &names[7], PRMpointOneDefaults),

    PRM_Template(PRM_FLT_J, 1, &names[9], &PRMtolDefault),

    PRM_Template(PRM_TOGGLE, 1, &names[10], PRMzeroDefaults),

    PRM_Template(PRM_TOGGLE, 1, &names[18], PRMzeroDefaults),

    PRM_Template(PRM_INT_J, 1, &names[19], PRMzeroDefaults),

    PRM_Template(PRM_STRING, 1, &names[11],&PRMinGrpDefault),

    PRM_Template(PRM_STRING, 1, &names[12],&PRMonGrpDefault),

    PRM_Template(PRM_STRING, 1, &names[13],&PRMoutGrpDefault),

    PRM_Template(PRM_STRING, 1, &names[20],&PRMprefixDefault),

    PRM_Template(),
};

unsigned
SOP_PolyCrease::disableParms()
{
    unsigned changed = 0;

    changed  = enableParm("t", NEWT()*(EMBED()==2?0:1));

    changed += enableParm("topoffset", SIDERAILS()*(EMBED()==2?0:1));

    changed += enableParm("botoffset", SIDERAILS()*(EMBED()==2?0:1));

    changed += enableParm("disttype", SIDERAILS()*(EMBED()==2?0:1));

    changed += enableParm("snapendrails", SIDERAILS()*(EMBED()==2?0:1));

    changed += enableParm("agrp", OUTGRP()*PARTITION()*!SIDERAILS()*(EMBED()==2?0:1));
    changed += enableParm("ongrp", OUTGRP()*(EMBED()==2?0:1));
    changed += enableParm("bgrp", OUTGRP()*PARTITION()*!SIDERAILS()*(EMBED()==2?0:1));
    changed += enableParm("partition", OUTGRP()*CLOSEPATHS()*!SIDERAILS()*(EMBED()==2?0:1));

    changed += enableParm("ang", NEWANG()*(EMBED()==2?0:1));
    changed += enableParm("newang", 0*(EMBED()==2?0:1));

    changed += enableParm("outgrps",(EMBED()==2?0:1));

    changed += enableParm("newt",(EMBED()==2?0:1));

    changed += enableParm("closepaths",(EMBED()==2?0:1));
    changed += enableParm("pathtype",(EMBED()==2?0:1));

    changed += enableParm("seed",(EMBED()==2?0:1));
    changed += enableParm("siderails",(EMBED()==2?0:1));

    changed += enableParm("prefix",EMBED());

    //changed += enableParm("outgrps",!USEBIAS());

    return changed;
}

//input labels
const char *
SOP_PolyCrease::inputLabel( unsigned int input) const
{
 switch (input) {
   case 0:
     return "Input Geometry";
     break;
   case 1:
     return "Cutting Curves";
     break;
   default:
     return "WTF?";
     break;
 }
}

OP_Node *
SOP_PolyCrease::myConstructor(OP_Network *net, const char *name, OP_Operator *op)
{
    return new SOP_PolyCrease(net, name, op);
}

SOP_PolyCrease::SOP_PolyCrease(OP_Network *net, const char *name, OP_Operator *op)
  : SOP_Node(net, name, op)
{
  setHighlight(1);
}

SOP_PolyCrease::~SOP_PolyCrease() {}

bool
SOP_PolyCrease::findEdgeUsingPrim(float &u, GEO_Point *&pta, GEO_Point *&ptb, GEO_Primitive *&prim,
                                  UT_Vector3 ppt, GB_PrimitiveGroup *primlist, float thres)
{
  GEO_Point             *tpt0,*tpt1;
  UT_Vector3            vtot,vtop,atob;
  float                 testLength,edgeLength,testLengthb,shortestLength=MAXFLT;
  bool                  found = false;
  GEO_Primitive         *poly;
  GEO_Vertex            *vtx;

  FOR_ALL_GROUP_PRIMITIVES(gdp, primlist, poly)
  {
    //cout<<poly->getNum()<<endl;
    for (int j = 0; j < poly->getVertexCount(); j++)
    {
      vtx = &poly->getVertex(j);
      tpt0 = vtx->getPt();
      //cout<<vtx->getPt()<<endl;
      vtx = &poly->getVertex(SYSfmod((float)j+1,(float)poly->getVertexCount()));
      tpt1 = vtx->getPt();
      //cout<<vtx->getPt()<<endl;

      vtop = ppt-tpt0->getPos();
      vtot = tpt1->getPos()-tpt0->getPos();
      testLength = vtot.length();
      testLengthb = vtop.length();
      //cout<<testLength<<endl;
      vtop.normalize();
      vtot.normalize();
      if ((vtop-vtot).length()<thres && testLengthb<shortestLength)
      {
        found = true;
        edgeLength = testLength;
        shortestLength = testLengthb;
        pta = tpt0;
        ptb = tpt1;
        prim = poly;
      }
    }
  }

  if (found)
  {
    u = (ppt-pta->getPos()).length()/edgeLength;
    //cout<<"u "<<u<<endl;
    //cout<<"edge pt1 "<<pta->getNum()<<endl;
    //cout<<"edge pt2 "<<ptb->getNum()<<endl;
  }

  return found;

}


bool
SOP_PolyCrease::findEdgeUsingPoint(float &u, GEO_Point *&pta, GEO_Point *&ptb, GEO_Primitive *&prim,
                                   UT_Vector3 ppt, const GEO_Point *nearppt,
                                   const GB_PointGroup *ptgroup, GB_PointRefArray *prefarr, GB_PrimitiveGroup *primlist, float thres)
{
  UT_Vector4            vtot,vtop,atob;
  float                 testLength;
  bool                  found = false;
  float                 shortestEdge = MAXFLT;
  GEO_Point             *ptt;

  //check all connected points to find the edge that the point is on
  //for each connected point check if our test point is between it and the closest point
  //cout<<"connected points ";
  pta = (GEO_Point*)nearppt;

  vtop = ppt-pta->getPos();
  FOR_ALL_GROUP_POINTS(gdp, ptgroup, ptt)
  {
    vtot = ptt->getPos()-pta->getPos();
    testLength = vtot.length();
    //cout<<pta->getNum()<<" ";
    vtop.normalize();
    vtot.normalize();
    vtop[3]=0;
    vtot[3]=0;
    if ((vtop-vtot).length()<thres && testLength<shortestEdge)
    {
      connectedPrims->findPolysUsingEdge(*pta,*ptt,primlist,*prefarr);
      if (primlist->entries())
      {
        found = true;
        shortestEdge = testLength;
        ptb = ptt;
        prim = (gdp)->primitives().head(*(primlist));
      }
    }
  }
  //cout<<endl;
  //find the u point on the edge
  if (found)
    u = (ppt-pta->getPos()).length()/shortestEdge;
    //cout<<"u "<<u<<endl;
    //cout<<"edge pt1 "<<pta->getNum()<<endl;
    //cout<<"edge pt2 "<<ptb->getNum()<<endl;

  return found;
}


bool
SOP_PolyCrease::checkFlipEdge(GEO_Point *pta0, GEO_Point *ptb0, float t, GEO_Point *&pta, GEO_Point *&ptb, float &u)
{
  UT_Vector3             a,b,c,d,av,bv;
  GEO_Point             *ptmp;
  bool                  flip = false;

  a = pta->getPos()+(ptb->getPos()-pta->getPos())*0.5;
  b = pta0->getPos()+(ptb0->getPos()-pta0->getPos())*0.5;

  c = pta->getPos()+(ptb->getPos()-pta->getPos())*0.25;
  d = pta0->getPos()+(ptb0->getPos()-pta0->getPos())*0.25;

  float len1 = (b-a).length()+(d-c).length();
  float len2 = (d-a).length()+(c-b).length();

  if (len1 > len2) flip = true;
/*
  //dot product method this seems more prone to flipping than the above
  av = ptb->getPos()-pta->getPos();
  bv = ptb0->getPos()-pta0->getPos();

  av.normalize();
  bv.normalize();

  cout<<av.dot(bv)<<endl;
  cout<<av.dot(av)<<endl;

  if (av.dot(bv)>0.99) { flip = false; }
  else if (av.dot(bv)<-0.99) { flip = true; }
  else {

    a = pta->getPos()+(ptb->getPos()-pta->getPos())*0.5;
    b = pta0->getPos()+(ptb0->getPos()-pta0->getPos())*0.5;

    av = b-a;
    av.normalize();

    a = ptb->getPos()-pta->getPos();
    b = ptb0->getPos()-pta0->getPos();

    a.normalize();
    b.normalize();


    if (av.dot(b) < 0 && av.dot(a) < 0) flip = true;
    if (av.dot(b) > 0 && av.dot(a) > 0) flip = true;

  }
*/
  if (flip)
  {
     ptmp = pta;
     pta = ptb;
     ptb = ptmp;
     u = 1-u;
  }

  return flip;
}

void
SOP_PolyCrease::appendSplitLocs(float now, bool found, float u, GEO_Point *pta, GEO_Point *ptb, GEO_Primitive *prim )
{
  bool                  flip=false;
  int                   i,start,splitcount,scratchptr=-1,ptr,allptr;
  float                 edgeLength,toff;
  UT_Vector4            edgev;
  const GB_PointGroup   *nullpointgroup=NULL;

  //append the splitLocs and allSplitLocs
  if (found && u>=0 && u<=1)
  {
    //cout<<"u "<<u<<endl;
    //add to SplitLocPtrArray list
    ptr = splitLocs.append();

    splitLocs[ptr] = new GU_SplitLoc;
    //cout<<"ptr "<<ptr<<endl;

    bool properedge = false;
    if (ptr>0 && u>0 && u<1)
    {
      for(i=ptr-1;i>=0;i--)
          if (splitLocs[i]->myT > 0.001 && splitLocs[i]->myT < 1)
          {
            properedge = true;
            //cout<<"proper edge"<<endl;
            break;
          }
      if (properedge)
        flip = checkFlipEdge(splitLocs[i]->myP0, splitLocs[i]->myP1,splitLocs[i]->myT, pta, ptb,u);
    }
    //cout<<"p0 "<<pta->getNum()<<endl;
    //cout<<"p1 "<<ptb->getNum()<<endl;
    //cout<<"flip "<<flip<<endl;
    splitLocs[ptr]->myP0 = pta;
    splitLocs[ptr]->myP1 = ptb;
    if (USEBIAS() && u) u = BIAS(now);
    splitLocs[ptr]->myT = u;
    splitLocs[ptr]->myPoly = (GEO_PrimPoly*)prim;

    if (EMBED()==2)
    {
      //add splitLocs to the allSplitLocs array
      allptr = allSplitLocs.append();
      allSplitLocs[allptr] = new GU_SplitLoc;
      allSplitLocs[allptr]->myP0 = splitLocs[ptr]->myP0;
      allSplitLocs[allptr]->myP1 = splitLocs[ptr]->myP1;
      allSplitLocs[allptr]->myT = splitLocs[ptr]->myT;
      allSplitLocs[allptr]->myPoly = splitLocs[ptr]->myPoly;
      currentSplitCount++;
    }
  }
}

void
SOP_PolyCrease::createExtraSplitLocs(float now)
{
  bool                  flip=false;
  int                   i,start,splitcount,scratchptr,ptr,allptr;
  float                 edgeLength,toff;
  UT_Vector4            edgev;
  const GB_PointGroup   *nullpointgroup=NULL;

  if (CLOSEPATHS() && splitLocs.entries() > 0)
  {
    ptr = splitLocs.append();

    splitLocs[ptr] = new GU_SplitLoc;

    splitLocs[ptr]->myP0 = splitLocs[0]->myP0;
    splitLocs[ptr]->myP1 = splitLocs[0]->myP1;
    splitLocs[ptr]->myT = splitLocs[0]->myT;
    splitLocs[ptr]->myPoly = splitLocs[0]->myPoly;

  }

  //append siderails if required
  if (OUTGRP())
  {
    //split scratch gdp with first set of splitLocs
    scratchgdp = new GU_Detail();
    scratchgdp->merge(*gdp);
    //copy all the splitLocs from the gdp to the scratchgdp, except the first and last if endpoints is on
    if (USEBIAS()) start = 1;
    else  start = 0;
    for (i=start;i<splitLocs.entries()-start;i++)
    {
      scratchptr = scratchSplitLocs.append();
      scratchSplitLocs[scratchptr] = new GU_SplitLoc;
      scratchSplitLocs[scratchptr]->myP0 = scratchgdp->points()(splitLocs[i]->myP0->getNum());
      scratchSplitLocs[scratchptr]->myP1 = scratchgdp->points()(splitLocs[i]->myP1->getNum());
      scratchSplitLocs[scratchptr]->myT = splitLocs[i]->myT;
      scratchSplitLocs[scratchptr]->myPoly = (GEO_PrimPoly*)scratchgdp->primitives()(splitLocs[i]->myPoly->getNum());
    }
    //do the split on the scratchgdp and get an edge group of the result
    //we can then use this later to find the central points on the split
    scratchoutedges = createAdhocEdgeGroup(*scratchgdp, "tmpedgegroup2");
    scratchgdp->edgeSplit(scratchSplitLocs,false,true,scratchoutedges);
    scratchpointtree.build(scratchgdp,nullpointgroup);
    scratchConnectedPrims = new GEO_Closure(*scratchgdp);
  }

  ptr = 0;
  splitcount = splitLocs.entries();
  if (SIDERAILS() && OFFSET("topoffset",now)>0)
  {
    for(i=splitcount-1;i>=0;i--)
    {
      edgev = splitLocs[i]->myP0->getPos()-splitLocs[i]->myP1->getPos();
      edgeLength = edgev.length();
      toff = PARA()?OFFSET("topoffset",now):OFFSET("topoffset",now)/(edgeLength>0.001?edgeLength:0.001);
      toff = (i==splitcount-1 || i == 0)?toff*!ENDPOINTS():toff;
      if (splitLocs[i]->myT+toff <= 1 && edgeLength)
      {
        ptr = splitLocs.append();
        splitLocs[ptr] = new GU_SplitLoc;
        splitLocs[ptr]->myP0 = splitLocs[i]->myP0;
        splitLocs[ptr]->myP1 = splitLocs[i]->myP1;
        splitLocs[ptr]->myT = splitLocs[i]->myT+toff;
        splitLocs[ptr]->myPoly = splitLocs[i]->myPoly;
      }
    }
  }

  if (SIDERAILS() && OFFSET("botoffset",now)>0)
  {
    for(i=0;i<splitcount;i++)
    {
      edgev = splitLocs[i]->myP0->getPos()-splitLocs[i]->myP1->getPos();
      edgeLength = edgev.length();
      toff = PARA()?OFFSET("botoffset",now):OFFSET("botoffset",now)/(edgeLength>0.001?edgeLength:0.001);
      toff = (i==splitcount-1 || i == 0)?toff*!ENDPOINTS():toff;
      if (splitLocs[i]->myT-toff >= 0 && edgeLength)
      {
        ptr = splitLocs.append();
        splitLocs[ptr] = new GU_SplitLoc;
        splitLocs[ptr]->myP0 = splitLocs[i]->myP0;
        splitLocs[ptr]->myP1 = splitLocs[i]->myP1;
        splitLocs[ptr]->myT = splitLocs[i]->myT-toff;
        splitLocs[ptr]->myPoly = splitLocs[i]->myPoly;
      }
    }
  }
}

void
SOP_PolyCrease::makeGroups(GU_PointGroup &newGroup,float tol)
{
  const GEO_Point       *ppt,*nearppt;
  const GB_PointGroup   *pointgroup;
  GB_PointGroup         *endGroup;

  endGroup = createAdhocPointGroup(*gdp, "endgroup");

  if (!SIDERAILS())
  {
    pointgroup=connectedPrims->getPointClosure(*outedges);
    FOR_ALL_GROUP_POINTS(gdp, pointgroup, ppt)
      newGroup.group()->add((GB_Element*)ppt);
    if (!USEBIAS())
    {
      FOR_ALL_GPOINTS(refgdp, ppt)
      {
        nearppt = pointtree.findNearestPt(ppt->getPos(),tol*2);
        newGroup.group()->add((GEO_Point*)nearppt);
      }
    }
  }
  else if (SIDERAILS())
  {
    pointgroup=scratchConnectedPrims->getPointClosure(*scratchoutedges);
   //loop through the edge group in the scratch gdp
    //finding the matching points in the real gdp
    FOR_ALL_GROUP_POINTS(scratchgdp, pointgroup, ppt)
    {
        nearppt = pointtree.findNearestPt(ppt->getPos(),tol*2);
        newGroup.group()->add((GEO_Point*)nearppt);
    }
    if (!USEBIAS())
    {
      int i = 0;
      FOR_ALL_GPOINTS(refgdp, ppt)
      {
        if (ENDPOINTS()) {
          if (i<refgdp->points().entries()-1 && i>0) {
            nearppt = pointtree.findNearestPt(ppt->getPos(),tol*2);
            newGroup.group()->add((GEO_Point*)nearppt); }
          else {
            nearppt = pointtree.findNearestPt(ppt->getPos(),tol*2);
            endGroup->add((GEO_Point*)nearppt); }
        }
        else {
            nearppt = pointtree.findNearestPt(ppt->getPos(),tol*2);
            newGroup.group()->add((GEO_Point*)nearppt); }
        i++;
      }
      newGroup.boolean(GU_GROUP_BOOLOP_MINUS,(GB_Group*)newGroup.group(),false,(GB_Group*)endGroup,false);
    }
  }
  delete scratchConnectedPrims;
  delete scratchgdp;
  int test = destroyAdhocGroup(scratchoutedges);
  test = destroyAdhocGroup(endGroup);
  scratchSplitLocs.entries(0);
}

void
SOP_PolyCrease::huntGdp(float &u, GEO_Point *&pta, GEO_Point *&ptb, GEO_Primitive *&prim,
                        UT_Vector3 ppt, GB_PrimitiveGroup *primlist, GB_PointRefArray *prefarr, GB_PointGroup *ptgroup,
                        float tol, bool &found)
{
  const GEO_Point       *nearppt;
  fpreal                maxdist=MAXFLT;
  GU_PointGroup         *expandptgroup;
  GB_PointGroup         *closeptgroup,*tmpgroup;
  UT_BoundingBox        bbox(-1,-1,-1,1,1,1);
  UT_PtrArray<GEO_Point*> pointlist;

  closeptgroup = createAdhocPointGroup(*gdp, "tmpNearPt");
  tmpgroup = createAdhocPointGroup(*gdp, "tmpGroup");
 //find neareat point on input gdp
  nearppt = pointtree.findNearestPt(ppt,maxdist);
  //cout<<"found nearest"<<endl;
  //find all connected prims
  primlist->clearEntries();
  connectedPrims->findPrimsUsingPoint(nearppt, primlist, *prefarr);
  //cout<<"closest point "<<nearppt->getNum()<<endl;
  //cout<<primlist->entries()<<endl;

  if (primlist->entries())
  {
    //check if the nearppt and the ppt are in the same place
    //if they are we can use any connected prim and edge
    if ((ppt - nearppt->getPos()).length()<tol)
    {
      closeptgroup->clearEntries();
      //put the closest point in a group
      closeptgroup->add((GEO_Point*)nearppt);
      expandptgroup = new GU_PointGroup(gdp, closeptgroup);
      //expand the group by one edge
      *ptgroup = *closeptgroup;
      *tmpgroup = *closeptgroup;
      expandptgroup->edgeDist(tmpgroup, 1);
      *ptgroup = *(GB_PointGroup*)expandptgroup->group();

      //cout<<"single point"<<endl;
      found = true;
      u = 0;
      pta = (GEO_Point*)nearppt;
      //ptb = (GEO_Point*)nearppt;
      ptb = (gdp)->points().head(*(ptgroup));
      prim = (gdp)->primitives().head(*(primlist));
      //cout<<"edge pt1 "<<pta->getNum()<<endl;
    }
    else
    {
      //cout<<"normal search"<<endl;
      found = findEdgeUsingPrim(u, pta, ptb, prim, ppt, primlist, tol);
      //if not found check a bigger neighourhood
      //and don't assume closest point is on the edge
      if (!found)
      {
        //cout<<"bigger search"<<endl;
        //make the radius a percentage of the bounding size
        //find all the points in that radius
        int test = gdp->getPointBBox(&bbox);
        float rad = bbox.sizeMax()/10;

        test = pointtree.findAllClosePt(ppt, rad, pointlist);

        //convert pointlist to a real group
        ptgroup->clearEntries();
        for(int i=0;i<pointlist.entries();i++)
          ptgroup->add(*pointlist[i]);

        //convert to primgroup
        primlist = connectedPrims->getPrimitiveClosure(*ptgroup);

        if (primlist->entries())
          found = findEdgeUsingPrim(u, pta, ptb, prim, ppt, primlist, tol);
      }
      //if still not found check the whole gdp
      if (!found)
      {
        //cout<<"total search"<<endl;
        //put all the points in a group
        ptgroup->clearEntries();

        FOR_ALL_PRIMITIVES(gdp, prim)
          primlist->add(prim);

        if (primlist->entries())
          found = findEdgeUsingPrim(u, pta, ptb, prim, ppt, primlist, tol);
      }
    }
  }
  int test = destroyAdhocGroup(closeptgroup);
  test = destroyAdhocGroup(tmpgroup);
}

OP_ERROR
SOP_PolyCrease::cookMySop(OP_Context &context)
{
  UT_Interrupt          *boss;
  const GB_PointGroup   *nullpointgroup=NULL,*pointgroup=NULL;
  GB_PointGroup         *outgroup;
  const GEO_Point       *nearppt,*ppt,*tppt;
  GEO_Point             *gppt;
  const UT_Vector3      vpt;
  UT_Vector3            testpoint;
  fpreal                maxdist=MAXFLT;
  GB_PrimitiveGroup     *primlist,*curvegroup;
  GEO_Point             *pta,*ptb;
  GEO_Primitive         *prim;
  const GEO_Primitive   *curve;
  float                 u=-1,shortestEdge,*splitattptr;
  GB_PointGroup         *ptgroup = NULL,*tmpgroup = NULL,*seedptgroup;
  const GB_PointGroup   *ptCurveGroup = NULL;
  GB_PointRefArray      *prefarr;
  int                   ptr=-1,tptr,bptr,allptr=0,i,splitatt,splitattoff,
                        curveattoff,*curveattptr,test,splitcount,scratchptr=-1,start,detoff;
  bool                  found=false,foundAtLeastOne=false;
  GU_PointGroup         *expandptgroup;
  UT_Vector3            nml;
  UT_String             finalGroup,groupA,groupB;
  const GEO_Vertex      *vtx;
  UT_IntArray           intCountArr;
  UT_String             append = "",prefix="",attname;

  //check for user interrupts
  boss = UTgetInterrupt();

  boss->opStart("Creasing polygons");

  bool secondInput = true;
  if (lockInputs(context) >= UT_ERROR_ABORT) return error();
  if( (refgdp = (GU_Detail*)inputGeo(1,context)) == NULL )
  {
    refgdp = new GU_Detail();
    secondInput = false;
  }

  float now = context.myTime;
  float tol = TOL(now);

  duplicateSource(0, context);

  //set the point weight to 1 for the whole gdp otherwise you get split errors
  FOR_ALL_GPOINTS(gdp, gppt)
  {
    gppt->getPos().w()=1.0;
  }
  FOR_ALL_PRIMITIVES(gdp, prim)
  {
      if (prim->getPrimitiveId() & GEOTPSURF)
        ((GEO_TPSurf *)prim)->weights(0);
      else if (prim->getPrimitiveId() & GEOFACE)
        ((GEO_Face *)prim)->weights(0);
  }

  //if (error() < UT_ERROR_ABORT && cookInputGroups(context) < UT_ERROR_ABORT && grp_name != "")
  //{
  GEO_Closure curveclosure   (*refgdp);

  curvegroup = createAdhocPrimGroup(*refgdp, "tmpCurveGroup");

  splitLocs.entries(0);
  allSplitLocs.entries(0);
  scratchSplitLocs.entries(0);

  outGroup(finalGroup,"ongrp");
  outGroup(groupA,"agrp");
  outGroup(groupB,"bgrp");

  GU_PointGroup newGroup(finalGroup, gdp);

  pointtree.build(gdp,nullpointgroup);
  connectedPrims = new GEO_Closure(*gdp);
  currentSplitCount=0;

  if (EMBED()==1 && !secondInput)
  {
    //check for layercount detail attribute - this is created by the layer sop
    //by supporting layering I can allow for different operations to be applied to different sets of embedded splits
    //This doesn't seem to work due to some limitiations of the hdk, i'll reinstate if fixed
    //for now i've added a user controlled prefix to do the same thing.
    append = "";
    PREFIX(prefix);

    detoff = gdp->findAttrib("currentlayer", sizeof(int), GB_ATTRIB_INT);

    if (detoff > -1)
    {
      int *currentlayer = (int*)gdp->attribs().getAttribData(detoff);
      append.itoa(*currentlayer);
    }

    //try to extract embeded splits
    attname = prefix;
    attname += "splitsTC";
    attname += append;
    attname += "_";
    detoff = gdp->findAttrib(attname, sizeof(int), GB_ATTRIB_INT);
    int *totalCurves = (int*)gdp->attribs().getAttribData(detoff);
    int atterror = detoff;

    attname = prefix;
    attname += "splitsTP";
    attname += append;
    attname += "_";
    detoff = gdp->findAttrib(attname, sizeof(int), GB_ATTRIB_INT);
    int *totalSplits = (int*)gdp->attribs().getAttribData(detoff);
    atterror += detoff;

    if (atterror > -1)
    {
      attname = prefix;
      attname += "creases";
      attname += append;
      attname += "_";
      int creasesOffset = gdp->findAttrib(attname, sizeof(int)*(*totalCurves), GB_ATTRIB_INT);

      attname = prefix;
      attname += "splits";
      attname += append;
      attname += "_";
      int attdim = 3*(*totalSplits);
      int splitsOffset = gdp->findAttrib(attname, sizeof(float)*attdim, GB_ATTRIB_FLOAT);
      if (creasesOffset < 0 || splitsOffset < 0)
      {
        addError(SOP_ERR_INVALID_SRC , "input 1 does not have embedded splits");
        unlockInputs();
        return error();
      }
      else
      {
        //loop through all the attributes creating the splitLocs
        int *creasesAtt = (int*)gdp->attribs().getAttribData(creasesOffset);
        float *splitsAtt = (float*)gdp->attribs().getAttribData(splitsOffset);
        int offset = -1;
        for (int i=0;i<*totalCurves;i++)
        {
          foundAtLeastOne=false;
          primlist = createAdhocPrimGroup(*gdp, "tmpPrims");
          ptgroup = createAdhocPointGroup(*gdp, "tmpptgroup");

          prefarr = new GB_PointRefArray(gdp,NULL);

          for (int j=0;j<creasesAtt[i];j++)
          {
            //look for stored point in the input mesh
            UT_Vector3 pA = gdp->points()((int)splitsAtt[++offset])->getPos();
            UT_Vector3 pB = gdp->points()((int)splitsAtt[++offset])->getPos();
            u = splitsAtt[++offset];
            testpoint = pA + (pB-pA)*u;
            found = false;
            huntGdp(u, pta, ptb, prim, testpoint, primlist, prefarr, ptgroup, tol, found);
            if (found)
              foundAtLeastOne=true;

            //create the splitLoc for the current point
            appendSplitLocs(now, found, u, pta, ptb, prim);
          }//end foreach split

          //create the extra splits, dependent on user options
          if (foundAtLeastOne)
          {
            createExtraSplitLocs(now);

            //do the actual splitting
            outedges = createAdhocEdgeGroup(*gdp, "tmpedgegroup");
            gdp->edgeSplit(splitLocs,PTYPE(),true,outedges);
            select(*outedges,1,1);

            //update the pointtree
            pointtree.build(gdp,nullpointgroup);
            delete connectedPrims;
            connectedPrims = new GEO_Closure(*gdp);

            if (OUTGRP())
              makeGroups(newGroup,TOL(now));

          }
          splitLocs.entries(0);

          test = destroyAdhocGroup(primlist);
          test = destroyAdhocGroup(ptgroup);

          delete prefarr;
        }//end foreach curve
      }//end else not invalid attributes
    }
    //destroy the detail attributes that stored all the splits
    attname = prefix;
    attname += "splitsTC";
    attname += append;
    attname += "_";
    gdp->destroyAttrib(attname, sizeof(int), GB_ATTRIB_INT);

    attname = prefix;
    attname += "splitsTP";
    attname += append;
    attname += "_";
    gdp->destroyAttrib(attname, sizeof(int), GB_ATTRIB_INT);

    attname = prefix;
    attname += "creases";
    attname += append;
    attname += "_";
    gdp->destroyAttrib(attname, GB_ATTRIB_INT);

    attname = prefix;
    attname += "splits";
    attname += append;
    attname += "_";
    gdp->destroyAttrib(attname, GB_ATTRIB_FLOAT);
  }//endif EMBED()==1
  else
  {
    //Try to create splits or set them up to be embedded
    //for each prim in second input
    FOR_ALL_PRIMITIVES(refgdp, curve)
    {
      foundAtLeastOne=false;
      currentSplitCount=0;
      prefarr = new GB_PointRefArray(gdp,NULL);

      //GU_PointGroup addGroup("tmp", gdp);
      primlist = createAdhocPrimGroup(*gdp, "tmpPrims");
      ptgroup = createAdhocPointGroup(*gdp, "tmpptgroup");

      //find all the points in the current curve
      curvegroup->clearEntries();
      curvegroup->add((GB_Element*)curve);
      ptCurveGroup = curveclosure.getPointClosure(*curvegroup);

      //for all points in current curve
      FOR_ALL_GROUP_POINTS(refgdp, ptCurveGroup, ppt)
      {
        //TODO add clean up if we are aborting
        if (boss->opInterrupt())
        {
          destroyAdhocGroups(gdp);
          destroyAdhocGroups(refgdp);
          delete prefarr;
          delete connectedPrims;
          unlockInputs();
          return error();
        }

        found = false;
        //look for current curve point in the input mesh
        testpoint = ppt->getPos();
        huntGdp(u, pta, ptb, prim, testpoint, primlist, prefarr, ptgroup, tol, found);
        if (found)
          foundAtLeastOne = true;

        //create the splitLoc for the current point
        appendSplitLocs(now, found, u, pta, ptb, prim);
      }//end FOR_ALL_GROUP_POINTS points in each curve in input 2
      if (EMBED()==2)
      {
        //save the actual number of splits found
        ptr = intCountArr.append();
        intCountArr[ptr]=currentSplitCount;
      }
      if (!EMBED())
      {
        if (foundAtLeastOne)
        {
          createExtraSplitLocs(now);

          //do the actual splitting
          outedges = createAdhocEdgeGroup(*gdp, "tmpedgegroup");
          gdp->edgeSplit(splitLocs,PTYPE(),true,outedges);
          select(*outedges,1,1);

          //update the pointtree
          pointtree.build(gdp,nullpointgroup);
          delete connectedPrims;
          connectedPrims = new GEO_Closure(*gdp);

          if (OUTGRP())
            makeGroups(newGroup,TOL(now));
        }

      }

      splitLocs.entries(0);

      test = destroyAdhocGroup(primlist);
      test = destroyAdhocGroup(ptgroup);

      delete prefarr;
    }//end FOR_EACH_PRIMITIVE
  }//endif

  if (OUTGRP() && PARTITION() && CLOSEPATHS() && found && !SIDERAILS() && !EMBED())
  {
    //expand the group by one edge
    GU_PointGroup aGroup(groupA, gdp);
    GU_PointGroup bGroup(groupB, gdp);

    seedptgroup = createAdhocPointGroup(*gdp, "seedgroup");
    ptgroup = createAdhocPointGroup(*gdp, "tmpptgroup");
    tmpgroup = createAdhocPointGroup(*gdp, "tmp");

    seedptgroup->clearEntries();
    ptgroup->clearEntries();

    seedptgroup->add(SEED(now));
    expandptgroup = new GU_PointGroup(gdp, seedptgroup);
    *ptgroup = *seedptgroup;
    *tmpgroup = *seedptgroup;
    bool growing = true;
    int lastnumingroup=1;
    while(growing)
    {
      if (boss->opInterrupt())
            break;

      //grow ptgroup by 1 edge
      //remove oncrease group
      expandptgroup->edgeDist(tmpgroup, 1);
      *ptgroup = *(GB_PointGroup*)expandptgroup->group();
      aGroup.boolean(GU_GROUP_BOOLOP_MINUS,(GB_Group*)ptgroup,false,(GB_Group*)newGroup.group(),false);
      *tmpgroup = *(GB_PointGroup*)aGroup.group();
      if (tmpgroup->entries() == lastnumingroup) growing = false;
      else lastnumingroup = tmpgroup->entries();

    }

    bGroup.boolean(GU_GROUP_BOOLOP_NONE,(GB_Group*)aGroup.group(),true,(GB_Group*)newGroup.group(),false);
    delete expandptgroup;
    //cout<<ptgroup->entries()<<endl;
    destroyAdhocGroups(gdp);
  }

  if (EMBED()==2)
  {
    //NOTE: I don't need to store the poly because it is re-found on extraction

    //check for layercount detail attribute - this is created by the layer sop
    //by supporting layering I can allow for different operations to be applied to different sets of embedded splits
    //This doesn't seem to work due to some limitiations of the hdk, i'll reinstate if fixed
    //for now i've added a user controlled prefix to do the same thing.
    append = "";
    PREFIX(prefix);

    detoff = gdp->findAttrib("currentlayer", sizeof(int), GB_ATTRIB_INT);

    if (detoff > -1)
    {
      int *currentlayer = (int*)gdp->attribs().getAttribData(detoff);
      append.itoa(*currentlayer);
    }

    allptr = 0;

    //read allSplitLocs and store as detail attribute string
    //format is #totalcurves, #totalpoints, #curve, #points, p0, p1, t, poly, #curve, #points, p0.......

    //create the detail atts
    i = intCountArr.entries();
    //cout<<i<<endl;
    attname = prefix;
    attname += "splitsTC";
    attname += append;
    attname += "_";
    gdp->destroyAttrib(attname, sizeof(int), GB_ATTRIB_INT);
    splitatt = gdp->addAttrib(attname,sizeof(int),GB_ATTRIB_INT,&i);

    i = allSplitLocs.entries();
    attname = prefix;
    attname += "splitsTP";
    attname += append;
    attname += "_";
    gdp->destroyAttrib(attname, sizeof(int), GB_ATTRIB_INT);
    splitatt = gdp->addAttrib(attname,sizeof(int),GB_ATTRIB_INT,&i);

    attname = prefix;
    attname += "creases";
    attname += append;
    attname += "_";
    gdp->destroyAttrib(attname, GB_ATTRIB_INT);

    curveattoff = gdp->addAttrib(attname,sizeof(int)*intCountArr.entries(),GB_ATTRIB_INT,0);
    curveattptr = (int*)gdp->attribs().getAttribData(curveattoff);
    for(i=0;i<intCountArr.entries();i++)
      curveattptr[i]=intCountArr[i];

    attname = prefix;
    attname += "splits";
    attname += append;
    attname += "_";
    gdp->destroyAttrib(attname, GB_ATTRIB_FLOAT);
    splitattoff = gdp->addAttrib(attname,sizeof(float)*allSplitLocs.entries()*3,GB_ATTRIB_FLOAT,0);
    splitattptr = (float*)gdp->attribs().getAttribData(splitattoff);
    for(i=0;i<allSplitLocs.entries();i++)
    {
      splitattptr[i*3]=(allSplitLocs[i]->myP0->getNum());
      splitattptr[i*3+1]=(allSplitLocs[i]->myP1->getNum());
      splitattptr[i*3+2]=(allSplitLocs[i]->myT);
      //splitattptr[i*4+3]=(allSplitLocs[i]->myPoly->getNum());
    }
  }

  if (UPDATE())
  {
    test = gdp->findPointAttrib("N", GB_ATTRIB_VECTOR);
    if (test > -1)
        gdp->normal();
  }

  if (!OUTGRP() || EMBED()==2) newGroup.destroy();
  destroyAdhocGroups(refgdp);
  destroyAdhocGroups(gdp);

  //}

  boss->opEnd();

  unlockInputs();

  return error();
}
